import pandas as pd
import numpy as np
from scipy.stats import ttest_ind
from statsmodels.stats.multitest import multipletests

def ttest_vs_all(df: pd.DataFrame,
                   testcolname: str="Z Score",
                   groupbycols: list=["Genes", "Precursor.Id", "Compound"],
                   ) -> pd.DataFrame:
    """
    Perform t-tests on Z scores grouped by specified columns.

    This function groups the input DataFrame by the specified columns and
    performs independent t-tests (Welch's t-test) for the Z scores of each
    group against the Z scores of all other groups combined. The results,
    including the t-statistic and p-value for each group, are returned in
    a new DataFrame.

    Parameters:
    ----------
    df : pd.DataFrame
        The input DataFrame containing the data.
    zcolname : str, optional
        The name of the column containing Z scores. Default is "Z Score".
    groupbycols : list, optional
        A list of column names to group the DataFrame by. Default is
        ["Genes", "Precursor.Id", "Compound"].

    Returns:
    -------
    pd.DataFrame
        A DataFrame containing the group keys, t-statistics, and p-values
        for each group. The columns include:
        - Each of the grouping columns specified in `groupbycols`.
        - `t_stat`: The t-statistic for the t-test.
        - `p_value`: The p-value for the t-test.
    """
        
    results = []
    groups = df.groupby(groupbycols, observed=False)

    for group_tuple, group in groups:
        # Z scores for this group
        specific_z_scores = group[testcolname]

        # Z scores for all other groups
        overall_z_scores = df[~df.index.isin(group.index)][testcolname]

        # Perform t-test if both groups have sufficient data
        if len(specific_z_scores) > 1 and len(overall_z_scores) > 1:
            t_stat, p_value = ttest_ind(
                specific_z_scores, overall_z_scores, equal_var=False
            )
        else:
            t_stat, p_value = np.nan, np.nan  # Not enough data for the test

        # Collect group information and results
        result_row = dict(zip(groupbycols, group_tuple))
        result_row.update({'t_stat': t_stat, 'p_value': p_value})
        results.append(result_row)

    return pd.DataFrame(results)

def protein_ttest_vs_dmso(df: pd.DataFrame,
                          protein_col_name: str="Genes",
                          compound_col_name: str="Compound",
                          control_compound: str="DMSO",
                          testcolname: str="Abundance"):
    """
    Perform t-tests on values grouped by protein.

    This function groups the input DataFrame by the protein and 
    compound and performs independent t-tests (Welch's t-test) for the specified
    value under treatment conditions vs control. The results,
    including the t-statistic and p-value for each group, are returned in
    a new DataFrame.

    Parameters:
    ----------
    df : pd.DataFrame
        The input DataFrame containing the data.
    protein_col_name   : str, optional
        The name of the column containing protein name. Default is "Genes".
    comound_col_name   : str, optional
        The name of the column containing compound name. Default is "Compound".
    control_compound   : str, optional
        The name of the control compound. Default is "DMSO".
    testcolname : str, optional
        The name of the column containing the values to test.
        Default is "Abundance".

    Returns:
    -------
    pd.DataFrame
        A DataFrame containing the group keys, t-statistics, and p-values
        for each group. The columns include:
        - Each of the grouping columns for protein and compound.
        - `t_stat`: The t-statistic for the t-test.
        - `p_value`: The p-value for the t-test.
    """  
    results = []
    groups = df.groupby([protein_col_name, compound_col_name], observed=False)

    for group_tuple, group in groups:
        protein, compound = group_tuple

        if compound == control_compound:
            continue
            
        treatment_values = group[testcolname]
        control_values = df[
            (df[protein_col_name] == protein) & 
            (df[compound_col_name] == control_compound)
        ][testcolname]

        if len(treatment_values) >= 3 and len(control_values) >= 3:
            t_stat, p_value = ttest_ind(
                treatment_values, control_values, equal_var=False
            )
        else:
            t_stat, p_value = np.nan, np.nan
        
        result_row = {
            protein_col_name: protein,
            compound_col_name: compound,
            "t_stat": t_stat,
            "p_value": p_value
        }
        results.append(result_row)
        return pd.DataFrame(results)

def peptide_ttest_vs_dmso(df: pd.DataFrame,
                          protein_col_name: str="Genes",
                          peptide_col_name: str="Precursor.Id",
                          compound_col_name: str="Compound",
                          control_compound: str="DMSO",
                          testcolname: str="Abundance"):
    """
    Perform t-tests on values grouped by peptide.

    This function groups the input DataFrame by the protein, peptide, and 
    compound and performs independent t-tests (Welch's t-test) for the specified
    value under treatment conditions vs control. The results,
    including the t-statistic and p-value for each group, are returned in
    a new DataFrame.

    Parameters:
    ----------
    df : pd.DataFrame
        The input DataFrame containing the data.
    protein_col_name   : str, optional
        The name of the column containing protein name. Default is "Genes".
    peptide_col_name   : str, optional
        The name of the column containing peptide name.
        Default is "Precursor.Id".
    comound_col_name   : str, optional
        The name of the column containing compound name. Default is "Compound".
    control_compound   : str, optional
        The name of the control compound. Default is "DMSO".
    testcolname : str, optional
        The name of the column containing the values to test.
        Default is "Abundance".

    Returns:
    -------
    pd.DataFrame
        A DataFrame containing the group keys, t-statistics, and p-values
        for each group. The columns include:
        - Each of the grouping columns for protein, peptide, and compound.
        - `t_stat`: The t-statistic for the t-test.
        - `p_value`: The p-value for the t-test.
    """
    results = []
    groups = df.groupby(
        [protein_col_name, peptide_col_name, compound_col_name], observed=False
    )
    for group_tuple, group in groups:
        protein, peptide, compound = group_tuple

        if compound == control_compound:
            continue

        treatment_values = group[testcolname]
        control_values = df[
            (df[protein_col_name] == protein) &
            (df[peptide_col_name] == peptide) &
            (df[compound_col_name] == control_compound)
        ][testcolname]

        if len(treatment_values) >= 3 and len(control_values) >= 3:
            t_stat, p_value = ttest_ind(
                treatment_values, control_values, equal_var=False
            )
        else:
            t_stat, p_value = np.nan, np.nan

        result_row = {
            protein_col_name: protein,
            peptide_col_name: peptide,
            compound_col_name: compound,
            "t_stat": t_stat,
            "p_value": p_value
        }
        results.append(result_row)
        return pd.DataFrame(results)


def correct_pvalues(df: pd.DataFrame,
                    pcolname: str="p_value",
                    corrected_col_name: str="corrected_p_value",
                    method: str="fdr_bh",
                    dropna: bool=True) -> pd.DataFrame:
    """
    Apply multiple testing correction to p-values in a DataFrame.

    This function adjusts p-values for multiple comparisons using the
    Benjamini-Hochberg method (False Discovery Rate). The corrected p-values
    are added as a new column to the DataFrame.

    Parameters:
    ----------
    df : pd.DataFrame
        The input DataFrame containing p-values.
    pcolname : str, optional
        The name of the column containing the p-values. Default is "p_value".
    corrected_col_name : str, optional
        The name of the column to store the corrected p-values.
        Default is "corrected_p_value".
    method : str, optional  
        The method to use for multiple testing correction. Default is "fdr_bh".
    dropna : bool, optional
        Whether to drop rows with missing p-values before correction.
        Default is True. If False and missing values present, correction will
        raise an error.

    Returns:
    -------
    pd.DataFrame
        The input DataFrame with an additional column containing the corrected
        p-values.
    """
    if dropna:
        df = df[df[pcolname].notna()]
    p_values = df[pcolname]
    _, corrected_p_values, _, _ = multipletests(p_values, method=method)
    df[corrected_col_name] = corrected_p_values
    return df
